package rwth.i2.ltlrv.afastate.impl;

import java.util.Set;
import java.util.Map.Entry;

import rwth.i2.ltlrv.afastate.base.UnaryAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.INext;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesHashMap;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;
import rwth.i2.ltlrv.util.SetUtils;

/**
 * Next - Implements the <i>Next</i> formula constructor of LTL.
 * 
 * @author Eric Bodden
 */
public class Next extends UnaryAFAState implements INext {

    /**
     * Constructs a new formula with the given subformula.
     * @param subFormula the subformula
     */
    public Next(IFormula subFormula) {
        super(subFormula);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFinalStateInAFA() {
        //"Next" states are nonfinal because they impose a requirement/obligation on a
        //next state, which must therefor exist
        return false;
    }

    /**
	 * {@inheritDoc}
	 */
    @Override
    @SuppressWarnings("unused")
    public void doTransition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result) {
        //only subformulae implementing IAFAState allowed at this stage
		assert (this.subformula instanceof IAFAState);
        
        //filter the binding: we only specialize with those bindings that are contained in the current context	
		WeakValuesMap<String, Object> filtered = filterBinding(context, currentBinding);
		
        //as next state, add the subformula with specialized bindings
		result.add( SetUtils.termClause(((IAFAState)subformula).specializeBindings(filtered)) );
	}

    /**
     * Filters bindings for those which are contained in the given context.
     * @param context the current context
     * @param binding the original binding
     * @return the filtered binding
     */
    private WeakValuesMap<String, Object> filterBinding(Set<String> context, WeakValuesMap<String, Object> binding) {
        WeakValuesMap<String, Object> filtered = new WeakValuesHashMap<String, Object>();
		for (Entry<String,Object> entry : binding.entrySet()) {
			if(context.contains(entry.getKey())) {
				filtered.put(entry.getKey(),entry.getValue());
			}
		}
        return filtered;
    }
    
    /**
     * Override cutting off the recursive search, since the <code>Next</code> operator
     * builds the boundary between the current and the next joinpoints. 
     * @see IAFAState#unboundVariablesAtCurrentJoinpoint(VariableKind, Set) 
     */
    @Override
    @SuppressWarnings("unused")
    //TODO still needed
    public void unboundVariablesAtCurrentJoinpoint(VariableKind variableKind, Set<String> result) {
        //The NEXT operator is the one who cuts off the search for unbound variables at the
        //*current* joinpoint. Thus here we abort the recursion.
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unused")
    public void providesPos(Set<String> result) {
		//the X operator separates temporal levels;
		//thus we do *not* take subformulae of a X operator
		//into account
	}
	
    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unused")
    public void providesNeg(Set<String> result) {
        //the X operator separates temporal levels;
        //thus we do *not* take subformulae of a X operator
        //into account
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    @SuppressWarnings("unused")
    //TODO why is this correct? update comment!
	public void requires(Set<String> result) {
		//the X operator separates temporal levels;
		//thus we do *not* take subformulae of a X operator
		//into account
	}
	
    /**
     * {@inheritDoc}
     */
    public IAFAState reconstruct(IFormula t) {
        return factory.X(t);
    }

    /**
     * {@inheritDoc}
     */
    public String symbol() {
        return "X";
    }
}
