package rwth.i2.ltlrv.afastate.impl;

import java.util.HashSet;
import java.util.Set;

import rwth.i2.ltlrv.afastate.base.CommutativeBinaryAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IOr;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;

/**
 * Or - Implements a disjunct.
 * 
 * @author Eric Bodden
 */
public class Or extends CommutativeBinaryAFAState implements IOr {

    /**
     * A disjunct of the two subformulae. 
     * @param subFormula1 left subformula
     * @param subFormula2 right subformula
     */
    public Or(IFormula subFormula1, IFormula subFormula2) {
        super(subFormula1,subFormula2);
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFinalStateInAFA() {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
		//this state is final if at least one subformula is final
        return ((IAFAState)this.subformula1).isFinalStateInAFA()
            || ((IAFAState)this.subformula2).isFinalStateInAFA(); 
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void doTransition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        //simply recurse, joining the results (remember the clause representation is in DNF)
        ((IAFAState)subformula1).transition(context,currentProps,currentBinding, result);
        ((IAFAState)subformula2).transition(context,currentProps,currentBinding, result);
	}

	/**
	 * {@inheritDoc}
	 */
	public void providesPos(Set<String> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        //recurse
        Set<String> res1 = new HashSet<String>();
        ((IAFAState)subformula1).providesPos(res1);
        Set<String> res2 = new HashSet<String>();
        ((IAFAState)subformula2).providesPos(res2);
        //intersect both results, cause a binding is certainly provided if it is provided
        //in both branches of the alternative
        for (String varName : res2) {
			if(res1.contains(varName)) {
				result.add(varName);
			}
		}
	}
    
    /**
     * {@inheritDoc}
     */
    public void providesNeg(Set<String> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        //join the results
        ((IAFAState)subformula1).providesNeg(result);
        ((IAFAState)subformula2).providesNeg(result);
    }
    
    /**
     * {@inheritDoc}
     */
    public IAFAState reconstruct(IFormula t1, IFormula t2) {
        return factory.Or(t1, t2);
    }
    
    /**
     * {@inheritDoc}
     */
    public String symbol() {
        return "||";
    }
    
}
