package rwth.i2.ltlrv.afastate.impl;

import java.util.Set;

import rwth.i2.ltlrv.afastate.base.BinaryAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.ITT;
import rwth.i2.ltlrv.afastate.interfaze.IUntil;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;

/**
 * Until - Implements the <i>Until</i> operator in LTL.
 * <i>Until</i> and <i>Release</i> subformulae are the essence of
 * an AFA representation of a LTL formula. They represent
 * the actual states of an AFA.s 
 * 
 * @author Eric Bodden
 * @see Release
 */

public class Until extends BinaryAFAState implements IUntil {

    /**
     * Constructs a new formula with the given subformulae.
     * @param subFormula1 left subformula
     * @param subFormula2 right subformula
     */
    public Until(IFormula subFormula1, IFormula subFormula2) {
        super(subFormula1,subFormula2);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public IAFAState negationNormalForm() {
        //recurse
        IUntil result = (IUntil)super.negationNormalForm();
        
        //remove duplicate operators
    	if (result.getSubformula2() instanceof IUntil) {
    		// phi U (phi U psi) == phi U psi
			IUntil s2 = (IUntil) result.getSubformula2();
			if(s2.getSubformula1().equals(subformula1)) {
		        return factory.Until(
		                s2.getSubformula1().negationNormalForm(),
		                s2.getSubformula2().negationNormalForm()
		                );
			}			
		}
    	
    	return result;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFinalStateInAFA() {
        //Until states are always fals in an AFA 
        return false;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void doTransition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState); 
        
        //delta(phi U psi) = delta(psi) || (delta(phi) && X(phi U psi) )
        factory.Or(
        		((IAFAState)subformula2),
                    factory.And(
                    		((IAFAState)subformula1),
                            factory.X(
                                    this
                            )
                    )
        ).transition(context,currentProps,currentBinding, result);
	}

	/**
	 * {@inheritDoc}
	 */
	public void providesPos(Set<String> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState); 
        
        //providesPos( phi U psi ) = providesPos( psi \/ (phi /\ X(phi U psi) )
        //                         = providesPos( psi ) intersect providesPos( phi )
        //                         = providesPos( phi \/ psi )
        factory.Or(subformula2,subformula1).providesPos(result);
	}  
    
    /**
     * {@inheritDoc}
     */
    public void providesNeg(Set<String> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState); 
        
        //providesNeg( phi U psi ) =  providesNeg( psi )
        ((IAFAState)subformula2).providesNeg(result);
    }
    
    /**
     * {@inheritDoc}
     */
    public IAFAState reconstruct(IFormula t1, IFormula t2) {
        return factory.Until(t1, t2);
    }

    /**
     * {@inheritDoc}
     */
    public String symbol() {
        return "U";
    }

    /**
     * {@inheritDoc}
     */
    @Override    
    public String toString() {      
        if(this.subformula1 instanceof ITT) {
            //render (TT U psi) as F(psi)
            return factory.F(this.subformula2).toString();
        } else {
            return super.toString();
        }
    }
}
