package rwth.i2.ltlrv.afastate.interfaze;

import rwth.i2.ltlrv.data.WeakValuesMap;

/**
 * IIfClosure - Interface for if-closures which are to be evaluated at runtime.
 * An if-closure evaluates a boolean expression under some given bindings.
 * This is used in order to evaluate if-pointcuts in the right context.
 * A typical example may be:<br><br>
 * 
 * <code><pre>
   new IIfClosure() {
                        public boolean satisfiedUnderBindings(
                                WeakValuesMap currentBindings) {
                            try {
                                final Integer j = (Integer) currentBindings
                                        .get("j");
                                final Integer i = (Integer) currentBindings
                                        .get("i");
                                try {
                                    return j != i;
                                } catch (java.lang.Exception ex) {
                                    throw new IIfClosure.UserCausedException(
                                            ex, "j != i");
                                }
                            } catch (java.lang.NullPointerException ex) {
                                return false;
                            } catch (java.lang.ClassCastException ex) {
                                return false;
                            }
                        }

                        public java.lang.String[] variableNames() {
                            return new java.lang.String[] { "j", "i" };
                        }

                        public java.lang.String toString() {
                            return "j != i";
                        }
                    }
   }
   </pre></code>
 * 
 * 
 * @author Eric Bodden
 */
public interface IIfClosure {

    /**
     * UserCausedException - An exception that wraps exceptions caused by
     * the evaluation of an IfClosure.
     * 
     * @author Eric Bodden
     */
    @SuppressWarnings("serial")
    public class UserCausedException extends RuntimeException {
        
        /** The expression that caused the exception. */
        private String ifExpression;
        
        /**
         * @param cause exception that was caused
         * @param ifExpression expression that caused the exception in String format
         */
        public UserCausedException(Throwable cause, String ifExpression) {
            super(cause);
            this.ifExpression = ifExpression;
        }

        /**
         * @return The expression that caused the exception.
         */
        public String getIfExpression() {
            return ifExpression;
        }
        
    }
    
    /**
     * Evaluates the given booelan expression under the given bindings.
     * @param currentBinding the binding to evaluate under
     * @return <code>true</code> if the expression is satisfied, <code>false</code> otherwise
     * @throws UserCausedException thrown if the boolean expression caused an exception;
     * the originally exception is wrapped in the {@link UserCausedException} and can be extracted
     * using {@link Throwable#getCause()}. The expression that caused the exception can be retrieved
     * by {@link UserCausedException#getIfExpression()} in String form.
     */
    public boolean satisfiedUnderBindings(WeakValuesMap<String, Object> currentBinding) throws UserCausedException;
    
    /**
     * Returns all variables required to evaluate this closure.
     * @return all variables required to evaluate this closure
     */
    public String[] variableNames();
    
    /**
     * Clients should implement this for debugging purposes.
     * @see Object#toString()
     */
    public String toString();
    
}