package rwth.i2.ltlrv.data;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

/**
 * A map supporting weak values.
 * Entries should disappear, when the value object is garbage collected. 
 * 
 * @author Eric Bodden
 * @param <K> type parameter for key type
 * @param <V> type parameter for value type
 */
public interface WeakValuesMap<K,V> extends Map<K,V>{
    
    /**
     * Locks the map, so that no modifications are allowed any more.
     * When a modifying method is called, a RuntimeException will be thrown. 
     */
    public void lock();
    
    /**
     * Returns if this map is locked.
     * When it is locked and a modifying method is called, a RuntimeException will be thrown. 
     * @return <code>true</code> if the map is locked
     */
    public boolean isLocked();
    
    /**
     * Provides access to an empty map. Used by the Backend.
     * 
     * @author bodden
     */
    public class Util {
    
    	    	
        private static final WeakValuesMap EMPTY_MAP = new WeakValuesMap() {

        	private boolean locked;
			public void lock() {
				locked = true;
			}

			public boolean isLocked() {
				return locked;
			}

			public int size() {
				return 0;
			}

			public boolean isEmpty() {
				return true;
			}

			@SuppressWarnings("unused")
            public boolean containsKey(Object key) {
				return false;
			}

            @SuppressWarnings("unused")
			public boolean containsValue(Object value) {
				return false;
			}

            @SuppressWarnings("unused")
			public Object get(Object key) {
				return null;
			}

            @SuppressWarnings("unused")
			public Object put(Object key, Object value) {
				throw new IllegalStateException("Put nothing into EMPTY_MAP!");
			}

            @SuppressWarnings("unused")
			public Object remove(Object key) {
				return null;
			}

            @SuppressWarnings("unused")
			public void putAll(Map t) {
				throw new IllegalStateException("Put nothing into EMPTY_MAP!");
			}

			public void clear() {
			}

			public Set keySet() {
				return Collections.EMPTY_SET;
			}

			public Collection values() {
				return Collections.EMPTY_LIST;
			}

			public Set entrySet() {
				return Collections.EMPTY_SET;
			}
        	
        };

		/**
         * @param <K> type parameter for key type
         * @param <V> type parameter for value type
		 * @return an unmodifyable, empty map
		 */
		@SuppressWarnings("unchecked")
        public static final <K,V> WeakValuesMap<K,V> emptyMap() {
        	return (WeakValuesMap<K,V>) EMPTY_MAP;
        }    	
    }

}
