package abc.ltl;

import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import polyglot.ast.NodeFactory;
import polyglot.frontend.BarrierPass;
import polyglot.frontend.Compiler;
import polyglot.frontend.CupParser;
import polyglot.frontend.FileSource;
import polyglot.frontend.GlobalBarrierPass;
import polyglot.frontend.Job;
import polyglot.frontend.Parser;
import polyglot.frontend.ParserPass;
import polyglot.frontend.Pass;
import polyglot.frontend.PrettyPrintPass;
import polyglot.frontend.VisitorPass;
import polyglot.lex.Lexer;
import polyglot.util.CodeWriter;
import polyglot.util.ErrorQueue;
import polyglot.visit.AmbiguityRemover;
import polyglot.visit.PrettyPrinter;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.parse.Lexer_c;
import abc.aspectj.visit.AJAmbiguityRemover;
import abc.aspectj.visit.AJTypeBuilder;
import abc.aspectj.visit.AspectNameCollector;
import abc.aspectj.visit.DeclareParentsAmbiguityRemover;
import abc.aspectj.visit.HierarchyBuilder;
import abc.aspectj.visit.InitClasses;
import abc.aspectj.visit.Jimplify;
import abc.aspectj.visit.NamePatternEvaluator;
import abc.aspectj.visit.NamePatternReevaluator;
import abc.aspectj.visit.NoSourceJob;
import abc.aspectj.visit.ParentDeclarer;
import abc.aspectj.visit.PatternTester;
import abc.ltl.ast.LTLNodeFactory_c;
import abc.ltl.formulaextraction.FormulaSource;
import abc.ltl.formulaextraction.Specification;
import abc.ltl.parse.Grm;
import abc.ltl.visit.override.CheckPackageNames;
import abc.ltl.visit.override.CollectJimplifyVisitor;
import abc.ltl.visit.transform.AdviceGeneration;
import abc.ltl.visit.transform.CleanUp;
import abc.ltl.visit.transform.ConstructorDeclaration;
import abc.ltl.visit.transform.GeneralFieldDecls;
import abc.ltl.visit.transform.ImportStatements;
import abc.ltl.visit.transform.MultiThreadingSecure;
import abc.ltl.visit.transform.PropositionLabels;
import abc.main.CompilerFailedException;

/**
 * Extension information for LTL extension.
 * @author Eric Bodden
 */
public class ExtensionInfo extends abc.aspectj.ExtensionInfo
{

    public static final polyglot.frontend.Pass.ID CONVERT_NEGATIONS = new polyglot.frontend.Pass.ID("convert-negations");
    public static final polyglot.frontend.Pass.ID IMPORTS = new polyglot.frontend.Pass.ID("imports");
    public static final polyglot.frontend.Pass.ID GENERAL_FIELD_DECLS = new polyglot.frontend.Pass.ID("general-field-declarations");
    public static final polyglot.frontend.Pass.ID PROPOSITION_LABELS = new polyglot.frontend.Pass.ID("proposition-labels");
    public static final polyglot.frontend.Pass.ID CONSTRUCTOR_DECL = new polyglot.frontend.Pass.ID("constructor-decl");
    public static final polyglot.frontend.Pass.ID PROPOSITION_ADVICE = new polyglot.frontend.Pass.ID("proposition-advice");
    public static final polyglot.frontend.Pass.ID CLEANUP = new polyglot.frontend.Pass.ID("cleanup");
	public static final polyglot.frontend.Pass.ID FORMULA_CONSTRUCTORS = new polyglot.frontend.Pass.ID("formula-constructors");
    public static final polyglot.frontend.Pass.ID FORMULA_IDS = new polyglot.frontend.Pass.ID("formula-ids");
    public static final polyglot.frontend.Pass.ID MULTI_THREADING_SECURE = new polyglot.frontend.Pass.ID("multi-threading-secure");
    public static final polyglot.frontend.Pass.ID BARRIER_1 = new polyglot.frontend.Pass.ID("barrier 1");
    public static final polyglot.frontend.Pass.ID BARRIER_2 = new polyglot.frontend.Pass.ID("barrier 2");
    public static final polyglot.frontend.Pass.ID LOAD_FORMULA_SOOT_CLASSES = new polyglot.frontend.Pass.ID("load-formula-soot-classes");

    static {
        // force Topics to load
        new Topics();
    }

    public ExtensionInfo(Collection jar_classes, Collection source_files)
    {
        super(jar_classes, source_files);
    }

    /** 
	 * @inheritDoc
	 */
    @Override
	public void initCompiler(Compiler comp) {
		super.initCompiler(comp);
        //load formulas into specification and add appropriate source units to
        //polyglot's jobs
        Specification.generateInitialSourceUnitsFromFormulas(this);
	}

    @Override
    public String compilerName() {
        return "ltl";
    }

    @Override
    public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
        Lexer lexer = new Lexer_c(reader, source.name(), eq);
        Grm grm = new Grm(lexer, ts, nf, eq);
        return new CupParser(grm, source, eq);
    }

    @Override
    protected NodeFactory createNodeFactory() {
        return new LTLNodeFactory_c();
    }
       
    /**
     * {@inheritDoc}
     */
    @Override
    public List passes(Job job) {

        //workaround: when no source files are given on the commandline, abc usually passes in a 
        //NoSourceJob to just execute all RunOnce passes. However, when we have to execute the full
        //compilation process, when formulae exist  
        if(job instanceof NoSourceJob && Specification.getFormulas().size()>0) {
            if (!compiler.compile(Collections.EMPTY_LIST)) {
                throw new RuntimeException(new CompilerFailedException("Compilation failed."));
            }
            return Collections.EMPTY_LIST;
        }
        
        List<Pass> l = new ArrayList<Pass>();
        l.add(new InitClasses(INIT_CLASSES, this, ts));

        l.add(new ParserPass(Pass.PARSE,compiler,job));

        l.add(new GlobalBarrierPass(BARRIER_1, job));
        
        if(!(job instanceof NoSourceJob) &&
                job.source() instanceof FormulaSource) {
            //l.add(new VisitorPass(CONVERT_NEGATIONS, job, new ConvertNegations((LTLNodeFactory)nf)));
            l.add(new VisitorPass(IMPORTS, job, new ImportStatements(nf)));
            l.add(new VisitorPass(GENERAL_FIELD_DECLS, job, new GeneralFieldDecls(nf)));
            l.add(new VisitorPass(PROPOSITION_LABELS, job, new PropositionLabels((AJNodeFactory)nf,ts)));
            l.add(new VisitorPass(CONSTRUCTOR_DECL, job, new ConstructorDeclaration(nf)));
            
            if(!(nf instanceof AJNodeFactory)) {
                throw new RuntimeException("Unexpected type of node factory (need AJNodeFactory): " + nf.getClass().getCanonicalName());
            }            
            l.add(new VisitorPass(PROPOSITION_ADVICE, job, new AdviceGeneration((AJNodeFactory)nf,ts)));
            l.add(new VisitorPass(MULTI_THREADING_SECURE, job, new MultiThreadingSecure(nf)));
            l.add(new VisitorPass(CLEANUP, job, new CleanUp()));
        }

        if(Debug.v().showGeneratedAspects) {
        	l.add(new PrettyPrintPass(INSPECT_AST,job,new CodeWriter(System.out,1000),new PrettyPrinter()));
        }

        l.add(new VisitorPass(Pass.BUILD_TYPES, job, new AJTypeBuilder(job, ts, nf)));

        l.add(new GlobalBarrierPass(Pass.BUILD_TYPES_ALL, job));
        l.add(new VisitorPass(Pass.CLEAN_SUPER, job,
                             new AmbiguityRemover(job, ts, nf, AmbiguityRemover.SUPER)));
        l.add(new BarrierPass(Pass.CLEAN_SUPER_ALL, job));
        ///
        passes_patterns_and_parents(l, job);
        passes_precedence_relation(l, job);
        //
        passes_disambiguate_signatures(l, job);
                

//        //
//        
        passes_add_members(l, job);
        passes_interface_ITDs(l, job);
        passes_disambiguate_all(l, job); 
        passes_fold_and_checkcode(l, job); 
        passes_saveAST(l, job);
        passes_mangle_names(l, job);
        passes_aspectj_transforms(l, job);
        
//        if(job.source() instanceof SyntheticSource) {
//            l.add(new VisitorPass(LOAD_FORMULA_SOOT_CLASSES,job, new FormulaSootClassLoader()));
//        }
        
        passes_jimple(l, job);
        
        return l;
    }

    @SuppressWarnings("unchecked")
    protected void passes_patterns_and_parents(List l, Job job)
    {
    	// Disambiguate inner/outer classes
    	l.add(new VisitorPass(CLEAN_CLASSES,job, new AJAmbiguityRemover(job,ts,nf,AmbiguityRemover.SIGNATURES)));
        // Disambiguate parents in declare parents
        l.add(new VisitorPass(CLEAN_DECLARE, job,
                              new DeclareParentsAmbiguityRemover(job, ts, nf))); 
        // Collect the full names of aspects
        l.add(new VisitorPass(COLLECT_ASPECT_NAMES, job, new AspectNameCollector(aspect_names)));

		//assure that synthetic sources are not added to the hierarchy
		//of weavable classes
//		if(!(job.source() instanceof SyntheticSource)) {
		// Build the internal hierarchy/package structure for the pattern matcher
        // The list of weavable classes is available after this pass
	        l.add(new VisitorPass(BUILD_HIERARCHY, job, new HierarchyBuilder(this)));
//		}
        l.add(new GlobalBarrierPass(HIERARCHY_BUILT, job));

        
        // Check that packages match directories
        //here we use a CUSTOM version
        l.add(new CheckPackageNames(CHECK_PACKAGE_NAMES,job));
        
        
        // Finds list of classes matched by each name pattern
        l.add(new VisitorPass(EVALUATE_PATTERNS, job, new NamePatternEvaluator(this)));
        if (abc.main.Debug.v().namePatternMatches) {
            // Print list of matched classes for each name pattern
            l.add(new VisitorPass(TEST_PATTERNS, job, new PatternTester(this)));
        }
        l.add(new GlobalBarrierPass(PATTERNS_EVALUATED, job));
        // Alter hierarchy according to declare parents (both internal pattern matching hierarchy
        // and Polyglot hierarchy)
        l.add(new VisitorPass(DECLARE_PARENTS, job, new ParentDeclarer(job, ts, nf, this)));
        l.add(new GlobalBarrierPass(PARENTS_DECLARED, job));
        // Finds list of classes matched by each name pattern, according to new hierarchy
        // but only for weavable classes.
        l.add(new NamePatternReevaluator(EVALUATE_PATTERNS_AGAIN));
        l.add(new GlobalBarrierPass(PATTERNS_EVALUATED_AGAIN, job));
    }

   @Override
   @SuppressWarnings("unchecked")
	protected void passes_jimple(List l, Job job)
    {
    	// set up map from top-level classes to asts for j2j
       //curstom CollectJimplifyVisitor! 
	   l.add(new VisitorPass(COLLECT_JIMPLIFY_CLASSES, job,
                              new CollectJimplifyVisitor(job, ts, nf, source_files, class_to_ast)));
        l.add(new GlobalBarrierPass(GOING_TO_JIMPLIFY, job));
        l.add(new Jimplify(JIMPLIFY, class_to_ast));
        l.add(new GlobalBarrierPass(JIMPLIFY_DONE, job));
    }
	
}
